/* Auto-Generated by RBGen version 0.1 on 2025-03-07T15:24:54Z */

#ifndef USR_SRC_RPM_BUILD_TREE_RB_XS_0_18_RBTREE_H
#define USR_SRC_RPM_BUILD_TREE_RB_XS_0_18_RBTREE_H

#include <stdbool.h>
#include <stddef.h>

/**
 * Red/Black tree node.  Nodes point to other nodes, so pointer math is
 * needed to reach the data paired with the node.
 */
typedef struct rbtree_node {
	struct rbtree_node *left, *right, *parent;
	size_t color: 1, count: 63;
} rbtree_node_t;

typedef int (*rbtree_compare_fn)(void *ctx, void *a, void *b);

#define rbtree_node_is_in_tree(node) ((bool) (node)->count)

/* Initialize a tree */
extern void rbtree_init_tree( rbtree_node_t *root_sentinel, rbtree_node_t *leaf_sentinel );

/* Returns previous node in conceptual sorted list of nodes, or NULL */
extern rbtree_node_t *rbtree_node_prev( rbtree_node_t *node );

/* Returns next node in conceptual sorted list of nodes, or NULL */
extern rbtree_node_t *rbtree_node_next( rbtree_node_t *node );

/* Returns right-most child of this node, or NULL */
extern rbtree_node_t *rbtree_node_right_leaf( rbtree_node_t *node );

/* Returns left-most child of this node, or NULL */
extern rbtree_node_t *rbtree_node_left_leaf( rbtree_node_t *node );

/* Returns the root-sentinel of the tree this node belongs to, or NULL */
extern rbtree_node_t *rbtree_node_rootsentinel( rbtree_node_t *node );

/* Add a node to a tree */
extern bool rbtree_node_insert( rbtree_node_t *hint, rbtree_node_t *node, rbtree_compare_fn cmp_fn, void *ctx, int cmp_pointer_ofs );
/* Manually alter tree assuming you've compared the keys already */
extern void rbtree_node_insert_before( rbtree_node_t *parent, rbtree_node_t *node );
extern void rbtree_node_insert_after( rbtree_node_t *parent, rbtree_node_t *node );

extern rbtree_node_t * rbtree_find_nearest( rbtree_node_t *node, void *goal,
	int(*cmp_fn)(void *ctx, void *a, void *b), void *ctx, int cmp_pointer_ofs,
	int *cmp_result );

/* Find a node matching or close to 'goal'.  Pass NULL for answers that aren't needed.
 * Returns true for an exact match (cmp==0) and false for anything else.
 */
extern bool rbtree_find_all( rbtree_node_t *node, void *goal, rbtree_compare_fn cmp_fn, void *ctx, int cmp_pointer_ofs,
	rbtree_node_t **result_first, rbtree_node_t **result_last, size_t *result_count );

/* In a tree with duplicate keys, find the left-most node with the same key as 'node', returning 'node' if no other. */
rbtree_node_t * rbtree_find_leftmost_samekey(rbtree_node_t *node, int(*compare)(void *ctx, void *a, void *b), void *ctx, int cmp_ptr_ofs);

/* In a tree with duplicate keys, find the right-most node with the same key as 'node', returning 'node' if no other. */
rbtree_node_t * rbtree_find_rightmost_samekey(rbtree_node_t *node, int(*compare)(void *ctx, void *a, void *b), void *ctx, int cmp_ptr_ofs);


/* Returns index of a node within the tree, as if the tree were a sorted array.
 * In other words, returns the number of nodes with a key less than this node.
 */
extern size_t rbtree_node_index( rbtree_node_t *node );

/* Returns Nth child of a node, or NULL */
extern rbtree_node_t *rbtree_node_child_at_index( rbtree_node_t *node, size_t index );


/* Remove a node from the tree */
extern void rbtree_node_prune( rbtree_node_t *node );


/* Remove all nodes from the tree, and optionally run a function on each */
extern void rbtree_clear( rbtree_node_t *root_sentinel, void (*destructor)(void *obj, void *opaque), int obj_ofs, void *opaque );

/* Check tree for internal validity */
#define RBTREE_INVALID_ROOT     1
#define RBTREE_INVALID_SENTINEL 2
#define RBTREE_INVALID_NODE     4
#define RBTREE_INVALID_COUNT    8
#define RBTREE_INVALID_COLOR   16
#define RBTREE_INVALID_ORDER   32
extern int rbtree_check_structure( rbtree_node_t *node, rbtree_compare_fn cmp_fn, void *ctx, int cmp_pointer_ofs );


#endif /* USR_SRC_RPM_BUILD_TREE_RB_XS_0_18_RBTREE_H */
